/* $LAAS: sdk.c 2008/07/04 17:36:54 mallet $ */

/* 
 * Copyright (c) 2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Jul  3 2008
 */

#define WINVER WindowsXP
#include <w32api.h>

#include <stdio.h>
#include <winsock2.h>
#include <ws2tcpip.h>

#include "EVaRT2.h"

void MyErrorMsgHandler(int iLevel, char *szMsg)
{
    char *szLevel;

    if (iLevel == VL_Debug)
    {
        szLevel = "Debug";
    }
    else
    if (iLevel == VL_Info)
    {
        szLevel = "Info";
    }
    else
    if (iLevel == VL_Warning)
    {
        szLevel = "Warning";
    }
    else
    if (iLevel == VL_Error)
    {
        szLevel = "Error";
    }

    printf("    %s: %s\n", szLevel, szMsg);
}

/* --- evas_evartinit ------------------------------------------------ */

int
evas_evartinit(void (*handler)(sFrameOfData *))
{
  char version[4];

  EVaRT2_SetVerbosityLevel(VL_Warning);

  EVaRT2_GetSdkVersion(version);
  printf("EVaRT SDK version %d.%d.%d\n",
	 version[1], version[2], version[3]);

#if 0
   EVaRT2_SetErrorMsgHandlerFunc(MyErrorMsgHandler);
#endif
  EVaRT2_SetDataHandlerFunc(handler);
  return 0;
}


/* --- evas_evartconnect --------------------------------------------- */

int
evas_evartconnect(char *myhost, char *evhost)
{
  sHostInfo hostinfo;
  int s;

  if (evhost) {
    /* if evhost was given, use that */
    s = EVaRT2_Initialize(myhost?myhost:evhost, evhost);
    if (s != RC_Okay) {
      fprintf(stderr, "could not initialize network\n");
      return -1;
    }

    s = EVaRT2_GetHostInfo(&hostinfo);
    if (s != RC_Okay || !hostinfo.bFoundHost) {
      fprintf(stderr, "could not connect to EVaRT\n");
      return -1;
    }

  } else {
    char host[1024], ip[16];
    struct addrinfo *evaddr, *a;
    int connected;
    char *r;

    s = gethostname(host, sizeof(host));
    if (s) { perror("cannot get local hostname"); return s; }

    /* retrieve all addresses of local host */
    s = getaddrinfo(host, NULL, NULL, &evaddr);
    if (s) { perror("cannot get localhost addresses"); return s; }

    connected = 0;
    for(a = evaddr; a && !connected; a = a->ai_next) switch (a->ai_family) {
      case AF_INET:
	r = inet_ntoa(((struct sockaddr_in *)a->ai_addr)->sin_addr);
	if (!r) break;

	strncpy(ip, r, sizeof(ip));
	fprintf(stderr, "trying to contact EVaRT on %s ... ", ip);
	fflush(stderr);

	s = EVaRT2_Initialize(myhost?myhost:ip, ip);
	if (s != RC_Okay) { fprintf(stderr, "failed\n"); break; }

	s = EVaRT2_GetHostInfo(&hostinfo);
	if (s != RC_Okay || !hostinfo.bFoundHost) {
	  fprintf(stderr, "failed\n"); break;
	}

	fprintf(stderr, "ok\n");
	connected = 1;
	break;

      default: break;
    }

    if (!connected) {
      fprintf(stderr, "could not connect to EVaRT\n");
      return -1;
    }
  }

  printf("%s Version %d.%d.%d at %d.%d.%d.%d (%s)\n",
	 hostinfo.szHostProgramName,
	 hostinfo.HostProgramVersion[1],
	 hostinfo.HostProgramVersion[2],
	 hostinfo.HostProgramVersion[3],

	 hostinfo.HostMachineAddress[0],
	 hostinfo.HostMachineAddress[1],
	 hostinfo.HostMachineAddress[2],
	 hostinfo.HostMachineAddress[3],
	 hostinfo.szHostMachineName);
  return 0;
}

