/* $LAAS: server.c 2009/10/05 17:34:31 mallet $ */

/*
 * Copyright (c) 2008-2009 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Jul  3 2008
 */

#define WINVER WindowsXP
#include <w32api.h>

#include <stdio.h>
#include <getopt.h>
#include <windows.h>
#include <winsock2.h>

#include "EVaRT2.h"
#include "evart-client.h"

#include "internal.h"


/* --- local data ---------------------------------------------------- */

static SOCKET servfd;

static struct {
  evas_msgtype_t type;
  struct sockaddr addr;
  char name[128];
} requests[64];

static int	evas_serverlisten(SOCKET sockfd);
static void	evas_datahandler(sFrameOfData* FrameOfData);
static int	evas_requestnew(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);
static int	evas_requestfree(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);
static int	evas_gettimeofacq(float delay, unsigned long *tv_sec,
			unsigned long *tv_usec);


/* --- main ---------------------------------------------------------- */

int
main(int argc, char* argv[])
{
  WSADATA supported;
  int s;

  /* options */
  static const char *usage =
    "usage: %s [--host evart_host] [--localhost this_host]\n"
    "\t\t\t[--port port_number]\n";
  static const struct option long_options[] = {
    { "host", 1, NULL, 'h' },
    { "localhost", 1, NULL, 'l' },
    { "port", 1, NULL, 'p' },
    { NULL, 0, NULL, 0 }
  };
  char *host = NULL;
  char *localhost = NULL;
  unsigned short port = 3615;

  while(1) {
    s = getopt_long(argc, argv, "h:l:p:", long_options, NULL);
    if (s == -1) break;

    switch (s) {
      case 'h':	host = optarg; break;
      case 'l': localhost = optarg; break;
      case 'p': port = atoi(optarg); break;

      case '?':
      default:
	printf(usage, argv[0]);
	exit(2);
    }
  }
  if (optind < argc) {
    printf(usage, argv[0]);
    exit(2);
  }

  /* setup sockets */
  s = WSAStartup(MAKEWORD(2, 2), &supported);
  if (s) {
    fprintf(stderr, "could not initialize sockets\n");
    return 2;
  }
  if (LOBYTE(supported.wVersion) != 2 || HIBYTE(supported.wVersion) != 2) {
    fprintf(stderr, "could not find a useable version of winsock.dll\n");
    WSACleanup();
    return 2;
  }

  /* connect to EVaRT */
  s = evas_evartinit(evas_datahandler);
  if (s) return s;

  s = evas_evartconnect(localhost, host);
  if (s) return s;

  /* setup server */
  servfd = evas_servercreate(port);
  if (servfd == INVALID_SOCKET) goto done;

  s = evas_serverlisten(servfd);

done:
    s = EVaRT2_Exit();
    printf("press any key to terminate...");
    s = getch();

    WSACleanup();
    return 0;
}


/* --- evas_servercreate --------------------------------------------- */

SOCKET
evas_servercreate(unsigned short port)
{
  SOCKET sockfd;
  struct sockaddr_in addr;
  int reuse;

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd == INVALID_SOCKET) return INVALID_SOCKET;

  /* set reuse addr option */
  reuse = 1;
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
		 (char *)&reuse, sizeof(reuse))) {
    perror("cannot set reuseaddr on server socket");
    goto error;
  }

  /* bind */
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);
  if (bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)))
    goto error;

  return sockfd;

error:
  fprintf(stderr, "cannot create socket server\n");
  closesocket(sockfd);
  return INVALID_SOCKET;
}


/* --- evas_serverlisten --------------------------------------------- */

static int
evas_serverlisten(SOCKET sockfd)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  struct sockaddr from;
  int fromlen = sizeof(from);
  void *response;
  int bytes;
  int i, s;

  while(1) {
    s = recvfrom(sockfd, buffer, sizeof(buffer), 0, &from, &fromlen);
    if (s <= 0) continue;

    switch(msg->type) {
      case EVAS_NONE: break;

      case EVAS_MSG:
	printf("* %s\n", msg->cmd);
	if (!strncmp(msg->cmd, "hello", 5)) {
	  msg->type = EVAS_MSG;
	  s = snprintf(msg->cmd, sizeof(msg->cmd), PACKAGE_STRING);
	  msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	  s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	}
	break;

      case EVAS_ACQUIRE:
	printf("* acquire %s\n", msg->cmd);
	if (msg->cmd[0] == '+') {
	  s = EVaRT2_Request("LiveMode", &response, &bytes);
	  s = (s != RC_Okay);
	} else {
	  s = EVaRT2_Request("PauseMode", &response, &bytes);
	  s = (s != RC_Okay);
	}
	msg->type = EVAS_ACQUIRE;
	s = snprintf(msg->cmd, sizeof(msg->cmd), "%d", s);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_RECORD:
	printf("* record '%s'\n", msg->cmd);
	if (msg->cmd[0] == '\0') {
	  s = EVaRT2_Request("StopRecording", &response, &bytes);
	  s = (s != RC_Okay);
	} else {
            char rqst[1024];

            snprintf(rqst, sizeof(rqst), "SetOutputName=%s", msg->cmd);
            s = EVaRT2_Request(rqst, &response, &bytes);
	    s = (s != RC_Okay);
	    if (!s) {
	      s = EVaRT2_Request("StartRecording", &response, &bytes);
	      s = (s != RC_Okay);
	    }
	}
	msg->type = EVAS_RECORD;
	s = snprintf(msg->cmd, sizeof(msg->cmd), "%d", s);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_FPS:
	printf("* fps\n");
	s = EVaRT2_Request("GetContextFrameRate", &response, &bytes);
	s = (s != RC_Okay);

	msg->type = EVAS_FPS;
	s = snprintf(
	  msg->cmd, sizeof(msg->cmd), "%c%f", s?'-':'+', *(float *)response);
	msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	break;

      case EVAS_BODY_LIST: {
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	printf("* listing %d bodies\n", bd->nBodyDefs);

	msg->len =
	  (char *)&msg->body_list.bodies[bd->nBodyDefs] - (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many bodies (%d)\n", bd->nBodyDefs);
	  break;
	}
	msg->type = EVAS_BODY_LIST;
	msg->body_list.nbodies = bd->nBodyDefs;
	for(i=0; i<bd->nBodyDefs; i++) {
	  strncpy(msg->body_list.bodies[i], bd->BodyDefs[i].szName,
		  sizeof(msg->body_list.bodies[i]));
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_MARKERS_LIST: {
	int body = atoi(msg->cmd);
	sBodyDefs *bd;

	bd = EVaRT2_GetBodyDefs();
	if (!bd) break;
	if (body < 0 || body >= bd->nBodyDefs) {
	  printf("* no such body\n");
	  break;
	}

	printf("* listing %d markers for body %d\n",
	       bd->BodyDefs[body].nMarkers, body);

	msg->len =
	  (char *)&msg->body_markers_list.markers[bd->BodyDefs[body].nMarkers] -
	  (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "* too many markers (%d)\n", bd->BodyDefs[body].nMarkers);
	  break;
	}
	msg->type = EVAS_BODY_MARKERS_LIST;
	msg->body_markers_list.index = body;
	strncpy(msg->body_markers_list.name, bd->BodyDefs[body].szName,
		sizeof(msg->body_markers_list.name));
	msg->body_markers_list.nmarkers = bd->BodyDefs[body].nMarkers;
	for(i=0; i<bd->BodyDefs[body].nMarkers; i++) {
	  strncpy(msg->body_markers_list.markers[i],
		  bd->BodyDefs[body].szMarkerNames[i],
		  sizeof(msg->body_markers_list.markers[i]));
	}
	sendto(sockfd, buffer, msg->len, 0, &from, fromlen);

	EVaRT2_FreeBodyDefs(bd);
	break;
      }

      case EVAS_BODY_MARKERS:
	switch(msg->cmd[0]) {
	  case '-':
	    s = evas_requestfree(EVAS_BODY_MARKERS, &from, msg->cmd+1);
	    break;

	  case '+':
	    s = evas_requestnew(EVAS_BODY_MARKERS, &from, msg->cmd+1);
	    break;

	  default:
	    s = evas_requestnew(EVAS_BODY_MARKERS, &from, msg->cmd);
	    break;
	}
	break;

      case EVAS_UNRECOGNIZED_MARKERS:
	switch(msg->cmd[0]) {
	  case '-':
	    s = evas_requestfree(EVAS_UNRECOGNIZED_MARKERS, &from, "");
	    break;

	  case '+':
	  default:
	    s = evas_requestnew(EVAS_UNRECOGNIZED_MARKERS, &from, "");
	    break;
	}
	break;
    }
  }

  return 0;
}


/* --- evas_datahandler ---------------------------------------------- */

static void
evas_datahandler(sFrameOfData *frame)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  int i, j, k;

  printf("%d / %d bodies       \r", frame->iFrame, frame->nBodies);
  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    switch(requests[i].type) {
      case EVAS_NONE: break;

      case EVAS_BODY_MARKERS:
	j = atoi(requests[i].name);
	if (j<0 || j>frame->nBodies) break;

	msg->len =
	  (char *)&msg->body_markers.markers[frame->BodyData[j].nMarkers] -
	  (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many markers (%d) for body %d:%s\n",
		  frame->BodyData[j].nMarkers,
		  j, frame->BodyData[j].szName);
	  continue;
	}
	msg->type = EVAS_BODY_MARKERS;
	msg->body_markers.iFrame = frame->iFrame;
	msg->body_markers.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
			  &msg->body_markers.tv_sec, &msg->body_markers.tv_usec);
	msg->body_markers.index = j;
	strncpy(msg->body_markers.name, frame->BodyData[j].szName,
		sizeof(msg->body_markers.name));
	msg->body_markers.nmarkers = frame->BodyData[j].nMarkers;
	for(k=0; k<msg->body_markers.nmarkers; k++) {
	  msg->body_markers.markers[k][0] = frame->BodyData[j].Markers[k][0];
	  msg->body_markers.markers[k][1] = frame->BodyData[j].Markers[k][1];
	  msg->body_markers.markers[k][2] = frame->BodyData[j].Markers[k][2];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;

      case EVAS_UNRECOGNIZED_MARKERS:
	msg->len =
	  (char *)&msg->body_markers.markers[frame->nUnidentifiedMarkers] -
	  (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many unidentified markers (%d)\n",
		  frame->nUnidentifiedMarkers);
	  continue;
	}
	msg->type = EVAS_UNRECOGNIZED_MARKERS;
	msg->unrecognized_markers.iFrame = frame->iFrame;
	msg->unrecognized_markers.fDelay = frame->fDelay;
	evas_gettimeofacq(frame->fDelay,
			  &msg->unrecognized_markers.tv_sec,
			  &msg->unrecognized_markers.tv_usec);
	msg->unrecognized_markers.nmarkers = frame->nUnidentifiedMarkers;
	for(k=0; k<msg->unrecognized_markers.nmarkers; k++) {
	  msg->unrecognized_markers.markers[k][0] = frame->UnidentifiedMarkers[k][0];
	  msg->unrecognized_markers.markers[k][1] = frame->UnidentifiedMarkers[k][1];
	  msg->unrecognized_markers.markers[k][2] = frame->UnidentifiedMarkers[k][2];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;
    }
}


/* --- evas_requestnew ----------------------------------------------- */

static int
evas_requestnew(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == EVAS_NONE) {
      requests[i].addr = *addr;
      if (name)
	snprintf(requests[i].name, sizeof(requests[i].name), "%s", name);
      else
	requests[i].name[0] = '\0';
      requests[i].type = type;
      return 0;
    }

  return -1;
}


/* --- evas_requestfree ---------------------------------------------- */

static int
evas_requestfree(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == type) {
      if (memcmp(addr, &requests[i].addr, sizeof(*addr))) continue;
      if (name)
	if (strcmp(name, requests[i].name)) continue;

      requests[i].type = EVAS_NONE;
      return 0;
    }

  return -1;
}


/* --- evas_gettimeofday --------------------------------------------------- */

static int
evas_gettimeofacq(float delay, unsigned long *tv_sec, unsigned long *tv_usec)
{
  static const unsigned long long epochms = 11644473600000000ULL;
  unsigned long long r;
  FILETIME ft;

  GetSystemTimeAsFileTime(&ft);
  r = ft.dwHighDateTime;
  r <<= 32;
  r |= ft.dwLowDateTime;

  /* converting file time to unix epoch */
  r /= 10;  /*convert into microseconds*/
  r -= epochms;

  /* substract delay */
  r -= (int)(delay * 1e6);
  *tv_sec = (long)(r / 1000000UL);
  *tv_usec = (long)(r % 1000000UL);
}
